/*  XMMS2 - X Music Multiplexer System
 *  Copyright (C) 2003-2006 XMMS2 Team
 *
 *  PLUGINS ARE NOT CONSIDERED TO BE DERIVED WORK !!!
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 */

package org.xmms2;

import java.io.FileDescriptor;
import java.io.FileInputStream;
import java.io.IOException;

import org.xmms2.xmms2bindings.Xmmsclient;
import org.xmms2.xmms2bindings.XmmsclientConstants;
import org.xmms2.xmms2bindings.SWIGTYPE_p_xmmsc_connection_St;
import org.xmms2.xmms2bindings.SWIGTYPE_p_xmmsc_result_St;

/**
 * Java way of using a mainloop. Just create a new Object and give a Callbacks-object, 
 * the mainloop does the rest :)
 */

public final class MainloopJava extends Thread{
	private SWIGTYPE_p_xmmsc_connection_St myConnection;
	private boolean running = false;

	/**
	 * constructor to set up the main loop and all needed callbacks
	 * 
	 * @param cb		Callbacks-object to call when something arrives
	 * @param conn		connection to listen on
	 */
	public MainloopJava(Callbacks cb, SWIGTYPE_p_xmmsc_connection_St conn){
		myConnection = conn;
		SpecialJNI.setENV(cb);
		running = true;
	}
	
	/**
	 * The run method - our little mainloop. First call SpecialJNI.setENV(this), which is a native method and sets "this"
	 * as GlobalRef for the callback-oject in c and sets the pointer jvm, which is needed to call java methods outside
	 * the JNI functions. Furthermore, init the callback-methods (set xmms2d's result_notifiers). While the file-descriptor 
	 * is valid, we will wait for new input and if something is available, call Xmmsclient.xmmsc_io_in_handle(myConnection)
	 * 
	 */
	public void run(){
		FileDescriptor fd = new FileDescriptor();
		SpecialJNI.getFD(fd, Xmmsclient.getPointerToConnection(myConnection));
		SpecialJNI.setupMainloop(this, Xmmsclient.getPointerToConnection(myConnection));
		FileInputStream in = new FileInputStream(fd);
		try {
			while (fd.valid() && running){
				Thread.sleep(40);
				if (in.available() > 0)
					Xmmsclient.xmmsc_io_in_handle(myConnection);
				if (Xmmsclient.xmmsc_io_want_out(myConnection) > 0)
					Xmmsclient.xmmsc_io_out_handle(myConnection);
			}
		} catch (InterruptedException e) {
			e.printStackTrace();
		} catch (IOException io){
			io.printStackTrace();
		}finally {
			try {
				in.close();
				Xmmsclient.xmmsc_io_disconnect(myConnection);
			} catch (Exception e){Xmmsclient.xmmsc_io_disconnect(myConnection);}
		}
	}
	
	/**
	 * detach from the mainloop
	 *
	 */
	public void spinDown(){
		running = false;
	}
	
	public void setConfigvalChangedCallback(int user_data){
		SWIGTYPE_p_xmmsc_result_St result = Xmmsclient.xmmsc_broadcast_configval_changed(myConnection);
		Xmmsclient.xmmsc_result_notifier_set(result, XmmsclientConstants.CALLBACK_CONFIGVAL_CHANGED, new int[]{user_data});
		Xmmsclient.xmmsc_result_unref(result);
	}
	
	public void setPlaylistCurrentPositionCallback(int user_data){
		SWIGTYPE_p_xmmsc_result_St result = Xmmsclient.xmmsc_broadcast_playlist_current_pos(myConnection);
		Xmmsclient.xmmsc_result_notifier_set(result, XmmsclientConstants.CALLBACK_PLAYLIST_CURRENT_POSITION, new int[]{user_data});
		Xmmsclient.xmmsc_result_unref(result);
	}
	
	public void setPlaybackCurrentIDCallback(int user_data){
		SWIGTYPE_p_xmmsc_result_St result = Xmmsclient.xmmsc_broadcast_playback_current_id(myConnection);
		Xmmsclient.xmmsc_result_notifier_set(result, XmmsclientConstants.CALLBACK_PLAYBACK_ID, new int[]{user_data});
		Xmmsclient.xmmsc_result_unref(result);
	}
	
	public void setPlaybackVolumeChangedCallback(int user_data){
		SWIGTYPE_p_xmmsc_result_St result = Xmmsclient.xmmsc_broadcast_playback_volume_changed(myConnection);
		Xmmsclient.xmmsc_result_notifier_set(result, XmmsclientConstants.CALLBACK_PLAYBACK_VOLUME_CHANGED, new int[]{user_data});
		Xmmsclient.xmmsc_result_unref(result);
	}
	
	public void setPlaybackStatusCallback(int user_data){
		SWIGTYPE_p_xmmsc_result_St result = Xmmsclient.xmmsc_broadcast_playback_status(myConnection);
		Xmmsclient.xmmsc_result_notifier_set(result, XmmsclientConstants.CALLBACK_PLAYBACK_STATUS, new int[]{user_data});
		Xmmsclient.xmmsc_result_unref(result);
	}
	
	public void setPlaylistChangedCallback(int user_data){
		SWIGTYPE_p_xmmsc_result_St result = Xmmsclient.xmmsc_broadcast_playlist_changed(myConnection);
		Xmmsclient.xmmsc_result_notifier_set(result, XmmsclientConstants.CALLBACK_PLAYLIST_CHANGED, new int[]{user_data});
		Xmmsclient.xmmsc_result_unref(result);
	}
	
	public void setMedialibEntryChangedCallback(int user_data){
		SWIGTYPE_p_xmmsc_result_St result = Xmmsclient.xmmsc_broadcast_medialib_entry_changed(myConnection);
		Xmmsclient.xmmsc_result_notifier_set(result, XmmsclientConstants.CALLBACK_MEDIALIB_ENTRY_CHANGED, new int[]{user_data});
		Xmmsclient.xmmsc_result_unref(result);
	}
	
	public void setMedialibEntryAddedCallback(int user_data){
		SWIGTYPE_p_xmmsc_result_St result = Xmmsclient.xmmsc_broadcast_medialib_entry_added(myConnection);
		Xmmsclient.xmmsc_result_notifier_set(result, XmmsclientConstants.CALLBACK_MEDIALIB_ENTRY_ADDED, new int[]{user_data});
		Xmmsclient.xmmsc_result_unref(result);
	}
	
	public void setMedialibPlaylistLoadedCallback(int user_data){
		SWIGTYPE_p_xmmsc_result_St result = Xmmsclient.xmmsc_broadcast_medialib_playlist_loaded(myConnection);
		Xmmsclient.xmmsc_result_notifier_set(result, XmmsclientConstants.CALLBACK_MEDIALIB_PLAYLIST_LOADED, new int[]{user_data});
		Xmmsclient.xmmsc_result_unref(result);
	}

	public void setMediainfoReaderStatusCallback(int user_data){
		SWIGTYPE_p_xmmsc_result_St result = Xmmsclient.xmmsc_broadcast_mediainfo_reader_status(myConnection);
		Xmmsclient.xmmsc_result_notifier_set(result, XmmsclientConstants.CALLBACK_MEDIAINFO_READER_STATUS, new int[]{user_data});
		Xmmsclient.xmmsc_result_unref(result);
	}

	public void setPlaybackPlaytimeSignal(int user_data){
		SWIGTYPE_p_xmmsc_result_St result = Xmmsclient.xmmsc_signal_playback_playtime(myConnection);
		Xmmsclient.xmmsc_result_notifier_set(result, XmmsclientConstants.SIGNAL_PLAYBACK_PLAYTIME, new int[]{user_data});
		Xmmsclient.xmmsc_result_unref(result);
	}
	
	public void setVisualizationDataSignal(int user_data){
		SWIGTYPE_p_xmmsc_result_St result = Xmmsclient.xmmsc_signal_visualisation_data(myConnection);
		Xmmsclient.xmmsc_result_notifier_set(result, XmmsclientConstants.SIGNAL_VISUALIZATION_DATA, new int[]{user_data});
		Xmmsclient.xmmsc_result_unref(result);
	}
	
	public void setMediareaderUnindexed(int user_data){
		SWIGTYPE_p_xmmsc_result_St result = Xmmsclient.xmmsc_signal_mediainfo_reader_unindexed(myConnection);
		Xmmsclient.xmmsc_result_notifier_set(result, XmmsclientConstants.SIGNAL_MEDIAREADER_UNINDEXED, new int[]{user_data});
		Xmmsclient.xmmsc_result_unref(result);
	}
	
	public void callbackIOWantOut(int val, int user_data){
		if (Xmmsclient.xmmsc_io_want_out(myConnection) > 0)
			Xmmsclient.xmmsc_io_out_handle(myConnection);
	}
}
